@@ -71,6 +71,30 @@ class Dense(Layer):
         self.outputs = np.dot(self.inputs, self.weights) + self.biases
         return np.dot(self.inputs, self.weights) + self.biases
     
+    def backward(self, dX, lr, momentum):
+        self.dY = dX
+        self.dW = np.dot(self.inputs.T, self.dY)
+        self.dB = np.sum(self.dY, axis=0, keepdims=True)
+        self.dX = np.dot(self.dY, self.weights.T)
+        
+        if momentum is not None:
+            weights = momentum * self.w_momentum - self.dW * lr
+            biases = momentum * self.b_momentum - self.dB * lr
+            
+            self.w_momentum = weights
+            self.b_momentum = biases
+            self.weights += weights
+            self.biases += biases
+            
+        elif momentum is None:
+            self.weights -= self.dW * lr
+            self.biases -= self.dB * lr
+        
+        self.dW *= np.zeros(self.dW.shape, dtype = np.float64)
+        self.dB *= np.zeros(self.dB.shape, dtype = np.float64)
+        
+        return self.dX
+    
 class ReLU(Layer):
     def __init__(self):
         super().__init__()
@@ -83,6 +107,11 @@ class ReLU(Layer):
     def forward(self):
         self.outputs = np.maximum(0,self.inputs)
         return np.maximum(0,self.inputs)
+    
+    def backward(self, dX, lr, momentum):
+        self.dY = dX
+        self.dX = self.dY * self.derivative()
+        return self.dX
 
 class Leaky_ReLU(Layer):
     def __init__(self):
@@ -97,6 +126,11 @@ class Leaky_ReLU(Layer):
         self.outputs = np.where(self.inputs > 0, self.inputs, self.inputs * 0.01)
         return np.where(self.inputs > 0, self.inputs, self.inputs * 0.01)
     
+    def backward(self, dX, lr, momentum):
+        self.dY = dX
+        self.dX = self.dY * self.derivative()
+        return self.dX
+    
 class Sigmoid(Layer):
     def __init__(self):
         super().__init__()
@@ -107,12 +141,23 @@ class Sigmoid(Layer):
     def forward(self):
         self.outputs = 1 / (1+np.exp(-self.inputs))
         return 1 / (1+np.exp(-self.inputs))
+    
+    def backward(self, dX, lr, momentum):
+        self.dY = dX
+        self.dX = self.dY * self.derivative()
+        return self.dX
 
 class Softmax(Layer):
     def __init__(self):
         super().__init__()
-        
-    def derivative(self, dE):
+    
+    def forward(self):
+        exp_values = np.exp(self.inputs - np.max(self.inputs, axis=1,keepdims=True))
+        probabilities = exp_values / np.sum(exp_values, axis=1,keepdims=True)
+        self.outputs = probabilities
+        return self.outputs
+    
+    def backward(self, dE):
         dA = np.empty_like(dE)
         for index, (single_output, single_dvalues) in enumerate(zip(self.outputs, dE)):
             single_output = single_output.reshape(-1, 1)
@@ -121,12 +166,6 @@ class Softmax(Layer):
             single_dvalues)
         return dA
     
-    def forward(self):
-        exp_values = np.exp(self.inputs - np.max(self.inputs, axis=1,keepdims=True))
-        probabilities = exp_values / np.sum(exp_values, axis=1,keepdims=True)
-        self.outputs = probabilities
-        return self.outputs
-    
 #%%Sequential 
 class Sequential():
     def __init__(self, train_inputs, train_targets, val_inputs = None, val_targets = None):
@@ -139,15 +178,15 @@ class Sequential():
         self.prediction = None
         self.train_layers = []
         self.val_layers = []
+        self.train_loss = None
+        self.val_loss = None
         self.train_loss_records = []
         self.val_loss_records = []
         self.train_accuracy_records = []
         self.val_accuracy_records =  []
-        self.__is_trained__ = False
         self.train_cost_fn = None
         self.val_cost_fn = None
-        self.train_loss = None
-        self.val_loss = None
+        self.__is_trained__ = False
         self.runtime = 0.0
     
     def add(self, ac_fn, n_in, n_out):
@@ -174,7 +213,6 @@ class Sequential():
                     break
             
     def forward(self):
-        from sys import exit
         self.train_layers[0].inputs = self.train_inputs
         for i in range(len(self.train_layers)):
             if i+1 == len(self.train_layers):
@@ -190,7 +228,6 @@ class Sequential():
                 self.val_layers[i+1].inputs = self.val_layers[i].forward()
                 
     def predict(self,x):
-        from sys import exit
         if self.__is_trained__ == True:
             self.train_layers[0].inputs = x
             for i in range(len(self.train_layers)):
@@ -200,42 +237,17 @@ class Sequential():
                     break
                 self.train_layers[i+1].inputs = self.train_layers[i].forward()
         else:
-            print("ERROR: Model not trained.")
-            exit()
+            raise RuntimeError("model not trained")
             
     def backward(self, lr, momentum):
-        from sys import exit
         if isinstance(self.train_layers[-1], Softmax):
-            self.dX = self.train_layers[-1].derivative(self.train_cost_fn.derivative(self.train_targets, self.train_outputs))
+            self.dX = self.train_layers[-1].backward(self.train_cost_fn.derivative(self.train_targets, self.train_outputs))
         else:
             self.dX = self.train_cost_fn.derivative(self.train_targets, self.train_outputs) * self.train_layers[-1].derivative()
         self.train_layers.reverse()
         for i in range(len(self.train_layers)):
-            if (isinstance(self.train_layers[i], Dense)):
-
-                self.dY = np.copy(self.dX)
-                self.dW = np.dot(self.train_layers[i].inputs.T, self.dY)
-                self.dB = np.sum(self.dY, axis=0, keepdims=True)
-                self.dX = np.dot(self.dY, self.train_layers[i].weights.T)
-                
-                if momentum is not None:
-                    weights = momentum * self.train_layers[i].w_momentum - self.dW * lr
-                    biases = momentum * self.train_layers[i].b_momentum - self.dB * lr
-                    
-                    self.train_layers[i].w_momentum = weights
-                    self.train_layers[i].b_momentum = biases
-                    self.train_layers[i].weights += weights
-                    self.train_layers[i].biases += biases
-                    
-                elif momentum is None:
-                    self.train_layers[i].weights -= self.dW * lr
-                    self.train_layers[i].biases -= self.dB * lr
-                
-                self.dW *= np.zeros(self.dW.shape, dtype = np.float64)
-                self.dB *= np.zeros(self.dB.shape, dtype = np.float64)
-            elif i > 0:
-                self.dY = np.copy(self.dX)
-                self.dX = self.dY * self.train_layers[i].derivative()
+            if i > 0:
+                self.dX = self.train_layers[i].backward(self.dX, lr, momentum)
         self.train_layers.reverse()
         if self.val_inputs is not None and self.val_targets is not None:
             for i in range(len(self.val_layers)):
@@ -244,7 +256,6 @@ class Sequential():
                     self.val_layers[i].biases = self.train_layers[i].biases
             
     def accuracy(self, targets, outputs):
-        from sys import exit
         accuracy = 1e-10
         for i in range(outputs.shape[1]):
             if i+2 > outputs.shape[1]:
@@ -255,7 +266,7 @@ class Sequential():
         
     def verbose(self,verbose,epoch,epochs,start):
         import time
-        from sys import exit
+        
         if verbose == 2:
             if epoch+1 < epochs:
                 if self.val_inputs is not None and self.val_targets is not None:
@@ -282,8 +293,7 @@ class Sequential():
         elif verbose == 0:
             return 
         else:
-            print("ERROR: Unknown value for verbose.")
-            exit()
+            raise ValueError("unknown value for verbose")
         
     def fit(self, cost_fn, epochs = 1000, lr = 0.1, momentum = None, verbose = 1, live_stats = False):
         self.__is_trained__ = True
@@ -292,7 +302,6 @@ class Sequential():
         
         from tqdm import tqdm
         import matplotlib.pyplot as plt
-        from sys import exit
         import time
         start = time.time()
         
@@ -387,7 +396,6 @@ class Sequential():
                 self.verbose(verbose, epoch, epochs, start)
         
     def results(self):
-        from sys import exit
         if self.__is_trained__ == True:
             import matplotlib.pyplot as plt
             
@@ -412,5 +420,4 @@ class Sequential():
             
             plt.show()
         else:
-            print("ERROR: Model not trained.")
-            exit()
\ No newline at end of file
+            raise RuntimeError("model not trained")
\ No newline at end of file
